/*
 * ChromosomeDistribution.cpp
 *
 * A haplotype frequency distribution based on a population size
 * Samples haplotypes from this distribution
 *
 *  Created on: Feb 25, 2011
 *      Author: jleigh
 */

#include <iostream>
#include <cstdlib>
using namespace std;

#include "ChromosomeDistribution.h"
#include "DomesticationException.h"

ChromosomeDistribution::ChromosomeDistribution(int popsize, const vector<Chromosome*> &haplotypes)
{
  _N = popsize;
  _nhaplotypes = haplotypes.size();
  for (vector<Chromosome*>::const_iterator cit = haplotypes.begin(); cit != haplotypes.end(); cit++)
    _htypes.push_back(*cit);

  _freqs = new double[_nhaplotypes];
  makerandomfreqs();

  _cumufreqranges = new int*[_nhaplotypes];
  int cumufreq = 0;
  for (int i = 0; i < _nhaplotypes; i++)
  {
    _cumufreqranges[i] = new int[2];
    _cumufreqranges[i][0] = cumufreq;
    cumufreq += (int)(_N * _freqs[i] + 0.5);
    _cumufreqranges[i][1] = cumufreq;
    //cout << "range: " << _cumufreqranges[i][0] << " : " <<_cumufreqranges[i][1] << endl;
  }

}

void ChromosomeDistribution::makerandomfreqs()
{
  int i, j, thisfreq, remaining = _N;

  double tmp, dN = (double)_N;

  for (i = 1; i < _nhaplotypes; i++)
  {
    thisfreq = rand() % (remaining + 1);
    _freqs[i-1] = thisfreq / dN;
    remaining -= thisfreq;
  }

  _freqs[_nhaplotypes-1] = remaining / dN;

  // Shuffle order
  for (i = 0; i < _nhaplotypes; i++)
  {
    j = rand() % (i + 1);
    tmp = _freqs[i];
    _freqs[i] = _freqs[j];
    _freqs[j] = tmp;
  }
}

const Chromosome* ChromosomeDistribution::samplehaplotype() const
{
  int u = rand() % _N;
  //cout << "u: " << u << endl;
  int idx = bsearchChromosome(u, 0, _nhaplotypes - 1);

  if (idx < 0)  return NULL;

  //cout << "found appropriate chromosome at index " << idx;

  //cout << ". Represents range " << _cumufreqranges[idx][0] << " to " << _cumufreqranges[idx][1] << endl;

  return _htypes.at(idx);
}

int ChromosomeDistribution::bsearchChromosome(int value, int first, int last) const
{
  if (first > last)  return -1;
  int middle = (last - first) / 2 + first;

  int lower = _cumufreqranges[middle][0];
  int upper = _cumufreqranges[middle][1];

  if (value >= lower && value < upper)  return middle;
  else if (value < lower)  return bsearchChromosome(value, first, middle - 1);
  else if (value >= upper)  return bsearchChromosome(value, middle + 1, last);

  else  return -1;
}

double ChromosomeDistribution::getfreq(int idx) const
{
  if (idx < 0 || idx >= _nhaplotypes)
    throw new DomesticationException("Invalid index for ChromosomeDistribution");

  return _freqs[idx];
}

const Chromosome* ChromosomeDistribution::gethaplotype(int idx) const
{
  if (idx < 0 || idx >= _nhaplotypes)
    throw new DomesticationException("Invalid index for ChromosomeDistribution");

  return _htypes.at(idx);
}


ChromosomeDistribution::~ChromosomeDistribution()
{
  delete[] _freqs;


  delete[] _cumufreqranges;
}
