#include <cstdlib>
#include <cstring>
#include <memory>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <sstream>
#include <unistd.h>
using namespace std;

#include "ChromosomeDistribution.h"
#include "Domestication.h"
#include "DomesticationException.h"

void usage()
{
  cerr << "Usage: domestication [-w wildpopsize] [-b bottlenecksize]"
      << " [-d domestsize]\n       [-B bottleneckduration]"
      <<" [-D domestduration] [-H hybridisedduration]\n       [-c nchrom]"
      << " [-p ploidy] [-l nloci] [-P npopulations]" << endl;

  cerr << "\nAll options take positive integers arguments:\n" << endl;
  cerr << "  -w wildpopsize\tSize of the wild population [default=10000]" << endl;
  cerr << "  -b bottlenecksize\tSize of the bottleneck population [default=20]" << endl;
  cerr << "  -d domestsize\t\tSize of domesticated populations [default=100]" << endl;
  cerr << "  -B bottleneckduration\tBottleneck time (generations) [default=10]" << endl;
  cerr << "  -D domestduration\tTime for domestication (generations) [default=20]" << endl;
  cerr << "  -H hybridisedduration\tTime after hybridisation (generations) [default=100]" << endl;
  cerr << "  -c nchrom\t\tHaploid number of chromosomes [default=20]" << endl;
  cerr << "  -p ploidy\t\tNumber of chromosome sets per genome [default=2]" << endl;
  cerr << "  -l nloci\t\tNumber of loci per chromosome [default=16]" << endl;
  cerr << "  -P npopulations\tNumber of domesticated populations [default=2]" << endl;
  cerr << "  -r precomb\tRecombination rate [default=0.01]" << endl;
  cerr << "  -s seed\tSet random number seed [default: use system time]" << endl;

}
void dogetopt(int argc, char **argv, int *hapsize, int *ploidy, int *nloci, int *npops,
    int *wildpopsize, int *bottlepopsize, int *domestpopsize, int *bottletime,
    int *domesttime, int *hybridtime, double *precomb, string *filename, int *seed)
{
  int value;
  double dvalue;
  int opt;

  while ((opt = getopt(argc, argv, "w:b:d:B:D:H:c:p:l:P:r:s:a:")) != -1)
  {
    //cout << "read opt " << (char)opt;
    //if (optarg != NULL)
    //  cout  << " with arg " << optarg << endl;
    //else
    //  cout << " with null arg" << endl;
    if (opt == '?')
    {
      cerr << "Invalid option or missing argument." << endl;
      usage();
      exit(1);
    }

    if (opt == 'r')
    {
      dvalue = atof(optarg);
      if (dvalue < 0 || dvalue > 1)
      {
        cerr << "Invalid value passed with -r:" << (char)opt << endl;
        usage();
        exit(1);
      }

      *precomb = dvalue;
      continue;
    }

    else if (opt == 'a')
    {
      filename->assign(optarg);
      continue;
    }

    value = atoi(optarg);
    //cout << "got value " << value << endl;
    if (value <= 0)
    {
      cerr << "Invalid value passed with option -" << (char)opt << endl;
      usage();
      exit(1);
    }

    switch (opt)
    {
      case 'c':
        *hapsize = value;
        break;
      case 'p':
        *ploidy = value;
        break;
      case 'l':
        *nloci = value;
        break;
      case 'P':
        *npops = value;
        break;
      case 'w':
        *wildpopsize = value;
        break;
      case 'b':
        *bottlepopsize = value;
        break;
      case 'd':
        *domestpopsize = value;
        break;
      case 'B':
        *bottletime = value;
        break;
      case 'D':
        *domesttime = value;
        break;
      case 'H':
        *hybridtime = value;
        break;
      case 's':
        *seed = value;
        break;
      default:
        cerr << "Invalid option: -" << opt << endl;
        usage();
        exit(0);
        break;

    }
  }
}

void makechromosometree(vector<Chromosome*> &tree, int nloci)
{
  Chromosome *chrm = new Chromosome(nloci);

  Chromosome *newchrm;
  vector<int> indices;
  for (int i = 0; i < nloci; i++)  indices.push_back(i);
  
  tree.push_back(chrm);
  int idx;
  
  vector<int>::iterator vit;
  
  for (int i = 1; i < nloci + 1; i++)
  {    
    idx = rand() % i;
    chrm = tree.at(idx);//rand() % i);
    newchrm = new Chromosome(*chrm);
    
    vit = indices.begin() + (rand() % indices.size());
    newchrm->setlocus(*vit, true);
    indices.erase(vit);
    tree.push_back(newchrm);
  }
}

void printgenome(Individual *indiv, const string &name)
{
  //cerr << "name: " << name << endl;

  cout.setf(ios::left, ios::adjustfield);

  const Chromosome *chrmptr;
  ostringstream oss;
  for (int i = 0; i < indiv->getPloidy(); i++)
  {
    oss.str("");
    oss << name << "_" << i;
    cout << setw(10) << oss.str() << ' ';
    for (int j = 0; j < indiv->getHaploidNumber(); j++)
    {
      chrmptr = indiv->getGenotype(j, i);
      for (int k = 0; k < chrmptr->getnloci(); k++)
        cout << chrmptr->getlocus(k);

    }
    cout << endl;
  }
}

void makesubpop(vector<Individual*> &individuals,  ChromosomeDistribution** chromdistros, int haplsize, int ploidy, int popsize)
{
  Individual *indiv;
  for (int i = 0; i < popsize; i++)
  {
    indiv = new Individual(ploidy, haplsize);
    for (int j = 0; j < haplsize; j++)
      for (int k = 0; k < ploidy; k++)
        indiv->setGenotype(j, k, chromdistros[j]->samplehaplotype());
    
    individuals.push_back(indiv);
  } 
}

/*
 * No mutation, no recombination
 */
void propagatesubpop(vector<Individual*> &individuals, int ngen, double precomb)
{
  vector<Individual*> nextgen;
  //Individual *indiv0, *indiv1;
  Individual *offspring;
  int idx0, idx1;
  //int tmp;
  int i, j; //, k;
  int ploidy, nchrom;

  //const Chromosome *chrmptr;

//  cout << "Parental population:" << endl;
//  for (vector<Individual*>::iterator it = individuals.begin(); it != individuals.end(); it++)
//  {
//  chrmptr= (*it)->getGenotype(0, 0);
//  for (j = 0; j < 16; j++)
//    cout << chrmptr->getlocus(j);
//  cout << "  ";
//
//  chrmptr = (*it)->getGenotype(0, 1);
//  for (j = 0; j < 16; j++)
//    cout << chrmptr->getlocus(j);
//  cout << endl;
//  }

  for (i = 0; i < ngen; i++)
  {
//    cout << "Generation " << i << endl;
    while (nextgen.size() < individuals.size())
    {
      idx0 = rand() % individuals.size();
      do
      {
        idx1 = rand() % individuals.size();
      }  while (idx1 == idx0);

      ploidy = individuals.at(idx0)->getPloidy();

      if (ploidy != 2)
        throw new DomesticationException("Subpopulation is not diploid.");

      nchrom = individuals.at(idx0)->getHaploidNumber();
      offspring = new Individual(ploidy, nchrom);

      for (j = 0; j < nchrom; j++)
      {


        //tmp = rand() % 2;
        //if (j == 0)  cout << "chroms from " << idx0 << "." << tmp;
        offspring->setGenotype(j, 0, individuals.at(idx0)->getGenotype(j, rand() % 2));
        //tmp = rand() % 2;
        //if (j == 0)  cout << " and " << idx1 << "." << tmp <<endl;
        offspring->setGenotype(j, 1, individuals.at(idx1)->getGenotype(j, rand() % 2));

        // decide whether to do recombination
        if (((double)rand()) / RAND_MAX < precomb)
        {
          bool tmp;
          Chromosome *chA = new Chromosome(*(offspring->getGenotype(j, 0)));
          Chromosome *chB = new Chromosome(*(offspring->getGenotype(j, 1)));
          int breakpoint = rand() % (chA->getnloci() - 1);
          /*cout << "doing a recombination at position " << breakpoint << " , chromosome " << j << endl;
          cout << "before:" << endl;
          for (int k = 0; k < chA->getnloci(); k++)
            cout << chA->getlocus(k);
          cout << endl;
          for (int k = 0; k < chB->getnloci(); k++)
            cout << chB->getlocus(k);
          cout << endl;*/

          for (int k = 0; k < breakpoint; k++)
          {
            tmp = chA->getlocus(k);
            chA->setlocus(k, chB->getlocus(k));
            chB->setlocus(k, tmp);
          }
          offspring->setGenotype(j, 0, chA);
          offspring->setGenotype(j, 1, chB);
          /*for (int k = 0; k < chA->getnloci(); k++)
            cout << chA->getlocus(k);
          cout << endl;
          for (int k = 0; k < chB->getnloci(); k++)
            cout << chB->getlocus(k);
          cout << "\n(and after)" << endl;*/
          delete chA;
          delete chB;
        }
      }

//      chrmptr = offspring->getGenotype(0, 0);
//      for (j = 0; j < 16; j++)
//        cout << chrmptr->getlocus(j);
//      cout << "  ";
//
//      chrmptr = offspring->getGenotype(0, 1);
//      for (j = 0; j < 16; j++)
//        cout << chrmptr->getlocus(j);
//      cout << endl;

      nextgen.push_back(offspring);

    }

    for (vector<Individual*>::iterator it = individuals.begin(); it != individuals.end(); it++)
      delete *it;
    individuals.assign(nextgen.begin(), nextgen.end());

    nextgen.clear();
  }
}

/*
 * This is probably dumb. Should instead expand over a generation
 */
void expandsubpop(vector<Individual*> &individuals, int newsize)
{
  int oldsize = individuals.size();
  if (oldsize >= newsize)
    throw new DomesticationException("expandsubpop was called to shrink or maintain a population.");

  for (int i = oldsize; i < newsize; i++)
    individuals.push_back(new Individual(*(individuals.at(rand() % oldsize))));
}

void mergesubpops(vector<Individual*> &mergedpop, vector<Individual*> *subpops, int npops, int mergedsize)
{
  //Individual *indiv;
  vector<Individual*> *pop;

  //for (int i = 0; i < mergedsize; i++)
  while (mergedpop.size() < mergedsize)
  {
    pop = &(subpops[rand() % npops]);
//    if (pop == &(subpops[0]))  cout << "drew from subpop 0, size "  << pop->size(); // << endl;
//    else if (pop == &(subpops[1]))  cout << "drew from subpop 1, size " << pop->size();//<< endl;
//    else cout << "unknown subpop." ;//endl;
    if (! pop->empty())
    {
//      int tmp = rand() % pop->size();
//      cout << ", indiv idx " << tmp << endl;
      vector<Individual*>::iterator iit = pop->begin() + (rand() % pop->size());
      mergedpop.push_back(*iit);
      pop->erase(iit);
    }

//    else cout << ", empty subpop." << endl;
    //indiv = pop->at(rand() % pop.size());
  }
}


//todo CHECK IF VECTORS ARE EMPTY IF THEY SHOULD BE [I don't remember what this means]
int main(int argc, char ** argv)
{
  //srand(time(NULL));
  
  /*
   * 1. Generate h chromosome trees (h = 20, haploid number)
   *    -> generate frequencies for both populations for each chromosome tree
   *    -> populationstructure class? weightedchrom class, with seq + freq? just add freq to Chromosome class?
   * 2. Generate n_b = 10 (user input?) individuals for each population
   *    -> for each chromosome tree, choose two chromosomes according to their frequencies
   *        -> use cumulative chrm indices: cumfreq[0] = count(chromosome[0]); cumfreq[1] = count(chromosome[1]) + cumfreq[0]..., then use a binary search
   * 3. Propagation: select two parents, randomly choose one of each chrom pair from each parent, make a child
   *     -> allow meiosis with some prob: select a random number to decide whether to cross over, then choose a breakpoint (uniformly)
   *     -> maybe skip meiosis at first
   *     -> expansion: propagate from parental generation to produce a larger population
   * 4. Admixture: select randomly from the two populations (expected 50/50) to produce a new population. The continue propagating.
   * 5. Analysis: produce distance matrix using Dice method ((b + c)/(2a + b + c), where a=1/1, b=1/0, c=0/1, based on 'or' combination within individuals)
   *      -> make NJ tree from distances
   *
   */

  /* Don't forget to take user input! */

  int hapsize = 20; // haploid number of chromosomes
  int ploidy = 2;
  int nloci = 16; // (segregating) loci per chromosome
  int npops = 2; // SHOULD BE 2!

  int N_w = 10000; // wild population size
  int N_b = 20; // bottleneck population size
  int N_d = 100; // domesticated population size

  int d_b = 10; // bottleneck duration
  int d_d = 20; // domestication duration
  int d_h = 100; // hybridisation duration
  
  double p_r = 0.01;

  int seed = time(NULL);
  string filename;

  dogetopt(argc, argv, &hapsize, &ploidy, &nloci, &npops, &N_w, &N_b, &N_d, &d_b, &d_d, &d_h, &p_r, &filename, &seed);
  

  srand(seed);
  
  int nsamps = 2 * (npops + npops + 1);
  
  // Not debugging code, don't comment out!
  cout << nsamps * 2 << ' ' << nloci * hapsize << endl;
  int i, j, k;//, k;
  //const Chromosome *chrmptr;

  // Set up chromosomes and frequency vectors
  vector<Chromosome*> *chromosomes = new vector<Chromosome*>[hapsize];

  // haplotype freq vector for each chromosome tree for each population
  ChromosomeDistribution ***haplotypes = new ChromosomeDistribution**[npops];

  for (i = 0; i < npops; i++)  haplotypes[i] = new ChromosomeDistribution*[hapsize];

  for (i = 0; i < hapsize; i++)
  {
    //cout << "Chromosome " << i << endl;
    makechromosometree(chromosomes[i], nloci);
    for (j = 0; j < npops; j++)
    {
      //popfreqs[j][i] = makerandomfreqs(nloci + 1, popsize);
      haplotypes[j][i] = new ChromosomeDistribution(N_w, chromosomes[i]);
    }

//    for (vector<Chromosome*>::iterator cit = chromosomes[i].begin(); cit != chromosomes[i].end(); cit++)
//
//      for (j = 0; j < chromosomes[i].size(); j++)
//      {
//        chrmptr = haplotypes[0][i]->gethaplotype(j);
//        for (int k = 0; k < nloci; k++)
//          cout << chrmptr->getlocus(k);
//
//        cout << " " << haplotypes[0][i]->getfreq(j) << " " <<  haplotypes[1][i]->getfreq(j);
//
//        cout << endl;
//      }
  }

//  i = 0;
//    for (j = 0; j < chromosomes[i].size(); j++)
//    {
//      chrmptr = haplotypes[0][i]->gethaplotype(j);
//      for (int k = 0; k < nloci; k++)
//        cout << chrmptr->getlocus(k);
//
//      cout << " " << haplotypes[0][i]->getfreq(j) << " " <<  haplotypes[1][i]->getfreq(j);
//
//      cout << endl;
//    }
  
  // Individual used only for sampling wild population
  Individual *indiv = new Individual(ploidy, hapsize);
  vector<Individual*>* subpops = new vector<Individual*>[npops];
  ostringstream oss;
  int idx0, idx1;

  for (i = 0; i < npops; i++)
  {
    for (j = 0; j < hapsize; j++)
      for (k = 0; k < ploidy; k++)
        indiv->setGenotype(j, k, haplotypes[i][j]->samplehaplotype());
    oss.str("");
    oss << "wild" << i << 0;
    //cerr << "oss str is now: " << oss.str() << endl;
    printgenome(indiv, oss.str());

    for (j = 0; j < hapsize; j++)
      for (k = 0; k < ploidy; k++)
        indiv->setGenotype(j, k, haplotypes[i][j]->samplehaplotype());
    oss.str("");
    oss << "wild" << i << 1;
    //cerr << "oss str is now: " << oss.str() << endl;
    printgenome(indiv, oss.str());


    makesubpop(subpops[i], haplotypes[i], hapsize, ploidy, N_b);
//    if (i == 0)
//    {
//      cout << "There are now " << subpops[i].size() << " individuals." << endl;
//        for (j = 0; j < hapsize; j++)
//        {
//          for (k = 0; k < ploidy; k++)
//          {
//            chrmptr = subpops[i].at(10)->getGenotype(j, k);
//            for (int l = 0; l < nloci; l++)
//              cout << chrmptr->getlocus(l);
//            cout << "   ";
//          }
//          cout << endl;
//        }
//    }
    propagatesubpop(subpops[i], d_b, p_r);
//    cout << "old size: " << subpops[i].size();
    expandsubpop(subpops[i], N_d);
//    cout << " new size: " << subpops[i].size() << endl;
    propagatesubpop(subpops[i], d_d, p_r);

    oss.str("");
    oss << "dom" << i << 0;
    idx0 = rand() % subpops[i].size();
    printgenome(subpops[i].at(idx0), oss.str());

    oss.str("");
    oss << "dom" << i << 1;
    do
    {
      idx1 = rand() % subpops[i].size();
    }  while (idx1 == idx0);
    printgenome(subpops[i].at(idx1), oss.str());

  }
  
  delete indiv;

  // todo look for memory leaks in simulations. Seems like lots of mem usage
  vector<Individual*> mergedpop;
  mergesubpops(mergedpop, subpops, npops, N_d);

  // Clean up old subpopulations
  for (i = 0; i < npops; i++)
    for (vector<Individual*>::iterator iit = subpops[i].begin(); iit < subpops[i].end(); iit++)
      delete *iit;
    
  delete[] subpops;

  //cout << "About to propagate merged population" << endl;
  propagatesubpop(mergedpop, d_h, p_r);


  idx0 = rand() % mergedpop.size();
  printgenome(mergedpop.at(idx0), "hyb0");

  do
  {
    idx1 = rand() % mergedpop.size();
  }  while (idx1 == idx0);
  printgenome(mergedpop.at(idx1), "hyb1");

  if (! filename.empty())
  {
    unsigned distcount = 0;
    double totaldist = 0; 
    fstream outfile(filename.c_str(), fstream::out|fstream::app);
    for (vector<Individual*>::iterator iit = mergedpop.begin(); iit < mergedpop.end(); iit++)
    {
      for (vector<Individual*>::iterator iit2(iit); iit2 != mergedpop.end(); iit2++)
      {
        unsigned a, b, c;
        a = b = c = 0;
        double dist = 0;
        bool char1, char2;

        for (unsigned i = 0; i < hapsize; i++)
        {
          const Chromosome * geno1a = (*iit)->getGenotype(i, 0);
          const Chromosome * geno1b = (*iit)->getGenotype(i, 1);
          const Chromosome * geno2a = (*iit2)->getGenotype(i, 0);
          const Chromosome * geno2b = (*iit2)->getGenotype(i, 1);

          for (unsigned j = 0; j < nloci; j++)
          {
            char1 = geno1a->getlocus(j) || geno1b->getlocus(j);
            char2 = geno2a->getlocus(j) || geno2b->getlocus(j);
            
            if (char1 && char2)  a += 1;
            else if (char1)  b += 1;
            else if (char2)  c += 1;
          }
        }
        if (2 * a + b + c)
          totaldist += (b + c)/((double)(2 * a + b + c));
        distcount ++;
      }
    }
    outfile << (totaldist/distcount) << endl;
    outfile.close();
  }

//  cout << "There are now " << subpop.size() << " individuals." << endl;
//  for (i = 0; i < hapsize; i++)
//  {
//    for (j = 0; j < ploidy; j++)
//    {
//      chrmptr = subpop.at(10)->getGenotype(i, j);
//      for (k = 0; k < nloci; k++)
//        cout << chrmptr->getlocus(k);
//      cout << "   ";
//    }
//    cout << endl;
//  }


  // Clean up populations
  for (vector<Individual*>::iterator iit = mergedpop.begin(); iit < mergedpop.end(); iit++)
    delete *iit;
  
  // Clean up frequency vectors
  for (i = 0; i < npops; i++)  delete[] haplotypes[i];

  delete[] haplotypes;

  // Clean up chromosomes
  for (i = 0; i < hapsize; i++)
  {
    for (vector<Chromosome*>::iterator cit = chromosomes[i].begin(); cit != chromosomes[i].end(); cit++)  delete *cit;
//    {
//      for (int j = 0; j < nloci; j++)
//        cout << (*cit)->getlocus(j);
//      cout << endl;
//      delete *cit;
//    }
  }

  delete[] chromosomes;


  // Set up frequencies
  // for each population, for each chromosome tree, for each genotype... a freq
  // pop1 has a double[nloci + 1][hapsize]
  
//  vector<Chromosome*> tree;
//
//  makechromosometree(tree, nloci);
//
//  for (vector<Chromosome*>::iterator cit = tree.begin(); cit < tree.end(); cit++)
//  {
//    for (int i = 0; i < nloci; i++)
//      cout << (*cit)->getlocus(i);
//    cout << endl;
//  }
//
//  int popsize = 10000;
//  double *freqs = makerandomfreqs(nloci + 1, popsize);
//
//  for (int i = 0; i <= nloci; i++)  cout << freqs[i] << endl;
//
//  delete freqs;
}
